/* wadsubs.c

  wad is a network daemon
   - when it starts up, it reads 1 file:
	wad.conf containing destinations to be monitored and their
	corresponding optimum buffer sizes along with other variables
	and stores this information in a linked list of Dest structures
	Send the SIGHUP signal to the process to make it re-read the 
	file and update the information--adds/changes.

   - notification via netlink:
     * the daemon issues a blocking read on a netlink socket waiting
	for an event:  0 means a tcp connection has started up
		       1 means a tcp connection has closed
	and its associated cid (connection id) for that connection
        Upon notification that a tcp connection has started up, 
	wad:
        + adds the new connection to a linked list of active cids
        + checks this connection against those it read from the
          configuration file (wad.conf) and:
            if there is a match
              if the connection is still in an established state
                if mode > 0, linux auto-tuning is set off and wad 
		  tunes the flows to this destination using values
		  from the config file 
	        Otherwise, linux auto-tuning is set on and wad 
		    does not tune the flows to this destination
              then it sets the send and receive buffers for all the
          active connections listed in wad.conf and sets the additive
          increase, max_ssthresh and scalable TCP parameters via 
	    the web100 interface
        As flows to a destination end, the buffer space is reallocated
          among the remaining flows
   - notification via /proc/web100:
    * (See description for WalkProcWeb100())
*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <string.h>
#include <signal.h>
#include <pthread.h>
#include <sys/time.h>
#include <math.h>
#include "web100.h"
#include "web100-int.h"

#define LOGFILE "waderr.log"
#define CONFIGFILE "wad.conf"
/*cids with these ports do not need to be adjusted*/
#define SSH_PORT 22
#define BWSERVER 7123  
#define HTTP_PORT 80
#define GRID_CTL 2811
#define MIB_CTL 8002
/*Connection Stage */
#define NOINTEREST 0
#define MONITOR    1
#define BUFFERSET  2
#define CIDCLOSED  3

/*GLOBAL VARIABLES*/
FILE *lf;
struct web100_var *CidState;
struct web100_var *CurMSS;
struct web100_var *CurMSStune;
struct web100_var *LimRwin;
struct web100_var *LimCwnd;
struct web100_var *WAD_AI;
struct web100_var *WAD_MD;
struct web100_var *WAD_IFQ;
struct web100_var *WAD_MaxSsthresh;
struct web100_var *WAD_Kaicnt;
struct web100_var *WAD_FloydAIMD;
struct web100_var *WAD_VegasGamma;
struct web100_var *WAD_VegasBeta;
struct web100_var *WAD_VegasAlpha;
struct web100_var *WAD_Vegas;
struct web100_var *CurCwnd;
struct web100_var *X_SBufMode;
struct web100_var *X_RBufMode;
u_int32_t AddrMask;

struct active_cid {
  int cid;
  unsigned int stage;
  unsigned int sbuf, rbuf;
  int up;
  int cwnd_ndx;
  struct active_cid *next_cid;
  struct active_cid *same_dest;
  struct Dest *cid_dest;
};
struct active_cid *CidRoot = NULL;

struct Dest {
  u_int32_t localadd, remoteadd;
  u_int32_t localmsk, remotemsk;
  u_int16_t localport, remoteport;
  unsigned int sbuf, rbuf;
  int mode;
  int wad_ai;
  int wad_md;
  int wad_maxssthresh;
  int wad_kaicnt;
  int wad_ifq;
  int wad_vegas;
  int wad_gamma;
  int wad_beta;
  int wad_alpha;
  int divide;
  int floyd;
  int curmss;
  int nconnects;
  struct active_cid *qfront;
  struct active_cid *qback;
  struct Dest *next_dest;
};
struct Dest *DestRoot = NULL;

int DynamicAIMD=0;

void read_config();
int ConvertAddress(char *, u_int32_t *, u_int32_t *);
void ConfigUpdate(struct Dest *);
void TuneConnections(struct web100_agent *, int);
void WalkProcWeb100(struct web100_agent *); 
struct active_cid *FindCid(int, int);
struct active_cid *AddCid(struct web100_connection *);
struct Dest *FilterByConnection(struct web100_connection *);
int CheckState(struct active_cid *, int);
int SetBuffers(struct active_cid *);
void *MonitorCurrentCwnd();

int err_sys(char *msg) {
  lf = fopen(LOGFILE, "a");
  if(lf==NULL) {
    fprintf(stderr, "%s\n", "err_sys: waderr.log open Error");
    exit(1);
  }
  fprintf(lf, "%s\n", msg);
  fclose(lf);
}

/* wad_init()
   sets up the necessary structures for monitoring connections and
   setting buffer sizes at the beginning
*/
struct web100_agent *wad_init() {
  struct web100_agent *agnt;
  struct web100_group *group;
  int i, retcode;
  pthread_t stid;
  struct Dest *dest;

  (int)AddrMask = -1;
  read_config(); 

/* Set up web100_agent, web100_group and web100_var structures */
  agnt=web100_attach(WEB100_AGENT_TYPE_LOCAL, NULL);
  if(!agnt) {
    fprintf(stderr,"wad_init: ERROR in creating agent!");
    exit(1);
  }
/* Get a pointer to the corresponding web100_var structure for 
   variables read/changed on a regular basis
*/
  if((group = web100_group_find(agnt, "read"))==NULL) {
    fprintf(stderr, "wad_init: can't locate /read\n");
    exit(1);
  }
  if((CidState = web100_var_find(group, "State"))==NULL) {
    fprintf(stderr, "wad_init: can't locate CidSstate\n");
    exit(1);
  }
  if((CurMSS = web100_var_find(group, "CurMSS"))==NULL) {
    fprintf(stderr, "wad_init: can't locate CurMSS\n");
    exit(1);
  }
  if((CurCwnd = web100_var_find(group, "CurCwnd"))==NULL) {
    fprintf(stderr, "wad_init: can't locate CurCwnd\n");
    exit(1);
  }
  if((group = web100_group_find(agnt, "tune"))==NULL) {
    fprintf(stderr, "wad_init: can't locate /tune\n");
    exit(1);
  }
  if((CurMSStune = web100_var_find(group, "CurMSS"))==NULL) {
    fprintf(stderr, "wad_init: can't locate CurMSS to tune\n");
    exit(1);
  }
  if((LimRwin = web100_var_find(group, "LimRwin"))==NULL) {
    fprintf(stderr, "wad_init: can't locate LimRwin\n");
    exit(1);
  }
  if((LimCwnd = web100_var_find(group, "LimCwnd"))==NULL) {
    fprintf(stderr, "wad_init: can't locate LimCwnd\n");
    exit(1);
  }
  if((X_RBufMode = web100_var_find(group, "X_RBufMode"))==NULL) {    fprintf(stderr, "wad_init: can't locate X_RBufMode\n");
    exit(1);
  }
  if((X_SBufMode = web100_var_find(group, "X_SBufMode"))==NULL) {    fprintf(stderr, "wad_init: can't locate X_SBufMode\n");
    exit(1);
  }
  if((WAD_AI = web100_var_find(group, "WAD_AI"))==NULL) {
    fprintf(stderr, "wad_init: can't locate WAD_AI\n");
    DynamicAIMD=0;
  }
  if((WAD_MD = web100_var_find(group, "WAD_MD"))==NULL) {
    fprintf(stderr, "wad_init: can't locate WAD_MD\n");
    DynamicAIMD=0;
  }
  if((WAD_FloydAIMD = web100_var_find(group, "WAD_FloydAIMD"))==NULL) {
    fprintf(stderr, "wad_init: can't locate WAD_FloydAIMD\n");
    DynamicAIMD=0;
  }
  if((WAD_IFQ = web100_var_find(group, "WAD_IFQ"))==NULL) {
    fprintf(stderr, "wad_init: can't locate WAD_IFQ\n");
    DynamicAIMD=0;
  }
  if((WAD_MaxSsthresh = web100_var_find(group, "WAD_MaxSsthresh"))==NULL) {
    fprintf(stderr, "wad_init: can't locate WAD_MaxSsthresh\n");
    DynamicAIMD=0;
  }
  if((WAD_Kaicnt = web100_var_find(group, "WAD_Kaicnt"))==NULL) {
    fprintf(stderr, "wad_init: can't locate WAD_Kaicnt\n");
  }
  if((WAD_Vegas = web100_var_find(group, "WAD_Vegas"))==NULL) {
    fprintf(stderr, "wad_init: can't locate WAD_Vegas\n");
  }
  else {
    if((WAD_VegasGamma = web100_var_find(group, "WAD_VegasGamma"))==NULL) {
      fprintf(stderr, "wad_init: can't locate WAD_VegasGamma\n");
    }
    if((WAD_VegasBeta = web100_var_find(group, "WAD_VegasBeta"))==NULL) {
      fprintf(stderr, "wad_init: can't locate WAD_VegasBeta\n");
    }
    if((WAD_VegasAlpha = web100_var_find(group, "WAD_VegasAlpha"))==NULL) {
      fprintf(stderr, "wad_init: can't locate WAD_VegasAlpha\n");
    }
  }
  if(DynamicAIMD) {
/* create a thread to monitor CurCwnd and change WAD_AI and WAD_MD
as needed*/
    retcode = pthread_create(&stid, NULL, MonitorCurrentCwnd, NULL);
    if(retcode < 0)
            fprintf(stderr, "wad_init: pthread_create error\n");
  }
  else
fprintf(stderr, "DynamicAIMD: %d!!\n", DynamicAIMD);
  return(agnt);
}

/* read_config()
   loads the connections from the wad.conf file into a linked list so 
   flows from these connections can be tuned.
*/
void read_config() {
  FILE *cf;
  double tmp1, tmp2;
  char laddr[32], raddr[32], line[1024], cmd[16];
  int lport, rport, records=0, errcode=0; 
  u_int32_t radd, rmask, ladd, lmask;
  unsigned int sbuf, rbuf;
  struct Dest newdest, *tmp;
  float kai;

/* kill -1 pid   will send the process the signal SIGHUP 
   to re-read the wad.conf file 
*/
      signal(SIGHUP, read_config);
  if((cf = fopen(CONFIGFILE, "r")) == NULL) {
    fprintf(stderr, "%s\n", "read_config: wad.conf File Open Error");
    exit(1);
  }
  memset(&newdest, 0, sizeof(struct Dest));
  strcpy(laddr, "0.0.0.0/32");
  strcpy(raddr, "0.0.0.0/32");
  while( fgets(line,sizeof(line),cf) ){
    if (strlen(line) < 5 || *line == '#') continue;  
    if(*line=='[') {
      if(records>0) {
/* Check to see if we have this Destination already 
   - if so, update it
   - if not, add it
*/
       if(!errcode) {
        ConfigUpdate(&newdest);
	memset(&newdest, 0, sizeof(struct Dest));
	strcpy(laddr, "0.0.0.0/32");
	strcpy(raddr, "0.0.0.0/32");
       }
      }
      errcode=0;
      records++;
      continue;
    }
    sscanf(line, "%s", cmd);
    if((strcmp(cmd, "src_addr:"))==0) {
      sscanf(line, "%s %s", cmd, laddr);
      if((errcode = ConvertAddress( laddr, &ladd, &lmask)) < 0)
	continue;
      newdest.localadd = ladd;
      newdest.localmsk = lmask;
    }
    else if((strcmp(cmd, "src_port:"))==0) 
      sscanf(line, "%s %d", cmd, &newdest.localport);
    else if((strcmp(cmd, "dst_addr:"))==0) {
      sscanf(line, "%s %s", cmd, raddr);
      if((errcode = ConvertAddress( raddr, &radd, &rmask)) < 0)
	continue;
      newdest.remoteadd = radd;
      newdest.remotemsk = rmask;
    }
    else if((strcmp(cmd, "dst_port:"))==0) 
      sscanf(line, "%s %d", cmd, &newdest.remoteport);
    else if((strcmp(cmd, "sndbuf:"))==0) {
      sscanf(line, "%s %d", cmd, &newdest.sbuf);
      if(newdest.sbuf <= 0) {
	fprintf(stderr, "Sndbuf for destination %s MUST contain a value!!\n", raddr);
	 exit(1);
      }
    }
    else if((strcmp(cmd, "rcvbuf:"))==0) {
      sscanf(line, "%s %d", cmd, &newdest.rbuf);
      if(newdest.rbuf <= 0) {
	fprintf(stderr, "Rcvbuf for destination %s MUST contain a value!!\n", raddr);
	exit(1);
      }
    }
    else if((strcmp(cmd, "maxssth:"))==0) 
      sscanf(line, "%s %d", cmd, &newdest.wad_maxssthresh);
    else if((strcmp(cmd, "kai:"))==0) {
      sscanf(line, "%s %f", cmd, &kai);
      if(kai > 0.0) 
	newdest.wad_kaicnt = (int)(1/kai);
      else
	newdest.wad_kaicnt = 0;
    }
    else if((strcmp(cmd, "wad_ifq:"))==0) 
      sscanf(line, "%s %d", cmd, &newdest.wad_ifq);
    else if((strcmp(cmd, "wadai:"))==0) {
      sscanf(line, "%s %lf", cmd, &tmp1);
      newdest.wad_ai = 8*tmp1;  /* shift 3 */
    }
    else if((strcmp(cmd, "wadmd:"))==0) {
      sscanf(line, "%s %lf", cmd, &tmp2);
      newdest.wad_md = 256*tmp2;  /* shift 8 */
    }
    else if((strcmp(cmd, "curmss:"))==0) 
      sscanf(line, "%s %d", cmd, &newdest.curmss);
    else if((strcmp(cmd, "mode:"))==0) {
      sscanf(line, "%s %d", cmd, &newdest.mode);
    }
    else if((strcmp(cmd, "wad_vegas:"))==0) {
      sscanf(line, "%s %d", cmd, &newdest.wad_vegas);
    }
    else if((strcmp(cmd, "gamma:"))==0) {
      sscanf(line, "%s %d", cmd, &newdest.wad_gamma);
    }
    else if((strcmp(cmd, "alpha:"))==0) {
      sscanf(line, "%s %d", cmd, &newdest.wad_alpha);
    }
    else if((strcmp(cmd, "beta:"))==0) {
      sscanf(line, "%s %d", cmd, &newdest.wad_beta);
    }
    else if((strcmp(cmd, "divide:"))==0) 
      sscanf(line, "%s %d", cmd, &newdest.divide);
    else if((strcmp(cmd, "floyd:"))==0) {
      sscanf(line, "%s %d", cmd, &newdest.floyd);
      if(newdest.floyd == 2)
        DynamicAIMD=1;
    }
  }
/* Check on last dest in config file */
  if(!errcode)
    ConfigUpdate(&newdest);
  fclose(cf);
/*
for(tmp=DestRoot; tmp; tmp=tmp->next_dest)
  fprintf(stderr, "%u-->", tmp->remoteadd);
fprintf(stderr, "\n");
*/
}

/* ConvertAddress()
   converts the config file address format into a u_int32_t,
   host-ordered, masked address.
*/
int ConvertAddress(char *addr, u_int32_t *daddr, u_int32_t *maddr) {  char *aptr, *tokptr, *saddr;
  u_int32_t mask_addr, destadd;
  int len, shiftm;
 
  aptr = strstr(addr, "/");
  if(aptr == NULL) {
    len = 32; 
  }
  else {
/* separate length from address string */
    tokptr = strtok(addr, "/");
    saddr = tokptr;
    tokptr = strtok(NULL, " ");
    len = atoi(tokptr);
    if(len < 1 || len > 32) {
      fprintf(stderr, "ERROR: Address length must be between 1 and 32
        Entry containing %s not added\n", addr);
      return -1;
    }
  } 
/* shift mask */
  shiftm = 32 - len;
  mask_addr = AddrMask << shiftm;
 
/*convert address to u_int32_t */
  destadd = inet_addr((char *)addr);
  destadd = ntohl(destadd);
 
/* apply mask and return values */
  *daddr = destadd & mask_addr;
  *maddr = mask_addr;
  return (0);
}

/* ConfigUpdate()
   walks the linked list of Dest structures looking for the
   destination address (raddr)  passed to it.  If one is found 
   update the data, otherwise add a new Dest Structure and link
   it to the back of the linked list. 
*/
void ConfigUpdate(struct Dest *destupd) { 
  unsigned int radd, ladd;
  struct Dest *dest, *olddest, *newdest;
  int found=0, chg=0;
  char *remadd;
  struct in_addr net_addr;

/* simple error check */
  if(destupd->remoteadd == 0 && destupd->remoteport == 0 && destupd->localport == 0) {
    fprintf(stderr, "ConfigAdd: The Remote address & both ports cannot ALL be 0!--entry NOT ADDED!\n");
    return;
  }

  found=0;
/* Look for this entry, if not found, add it! */
  for(dest=DestRoot; dest; dest=dest->next_dest) {
    if((destupd->remoteadd == dest->remoteadd) && (destupd->remoteport == dest->remoteport)) {
/* found it! */
      found=1; chg=0;
	if(dest->localadd != destupd->localadd) {
	  chg=1;
	  dest->localadd = destupd->localadd;
	}
	if(dest->localport != destupd->localport){
          chg=1; 
	  dest->localport = destupd->localport;
	}
	if(dest->sbuf != destupd->sbuf) {
	  chg=1;
	  dest->sbuf = destupd->sbuf;
	}
	if(dest->rbuf != destupd->rbuf) {
	  chg=1;
	  dest->rbuf = destupd->rbuf;
	}
        if(dest->mode != destupd->mode) {
	  chg=1;
	  dest->mode = destupd->mode;
	}
        if(dest->wad_ai != destupd->wad_ai) {
	  chg=1;
	  dest->wad_ai = destupd->wad_ai;
	}
        if(dest->wad_md != destupd->wad_md) {
	  chg=1;
	  dest->wad_md = destupd->wad_md;
	}
        if(dest->curmss != destupd->curmss) {
	  chg=1;
	  dest->curmss = destupd->curmss;
	}
        if(dest->wad_maxssthresh != destupd->wad_maxssthresh) {
	  chg=1;
	  dest->wad_maxssthresh = destupd->wad_maxssthresh;
	}
        if(dest->wad_kaicnt != destupd->wad_kaicnt) {
	  chg=1;
	  dest->wad_kaicnt = destupd->wad_kaicnt;
	}
        if(dest->wad_ifq != destupd->wad_ifq) {
	  chg=1;
	  dest->wad_ifq = destupd->wad_ifq;
	}
        if(dest->wad_vegas != destupd->wad_vegas) {
	  chg=1;
	  dest->wad_vegas = destupd->wad_vegas;
	}
        if(dest->wad_alpha != destupd->wad_alpha) {
	  chg=1;
	  dest->wad_alpha = destupd->wad_alpha;
	}
        if(dest->wad_beta != destupd->wad_beta) {
	  chg=1;
	  dest->wad_beta = destupd->wad_beta;
	}
        if(dest->wad_gamma != destupd->wad_gamma) {
	  chg=1;
	  dest->wad_gamma = destupd->wad_gamma;
	}
	if(dest->divide != destupd->divide) {
	  chg=1;
	  dest->divide = destupd->divide;
	}
        if(dest->floyd != destupd->floyd) {
          chg=1;
          dest->floyd = destupd->floyd;
        }
	if(chg) {
  net_addr.s_addr = htonl(dest->remoteadd);
  remadd = inet_ntoa(net_addr);
  fprintf(stderr, "%s %s %d sbuf: %u rbuf: %u mode: %d wad_ai: %d wad_md: %d, curmss: %d wad_maxssthresh: %d wad_kaicnt %d wad_ifq: %d divide: %d floyd: %d, vegas: %d, alpha: %d, beta: %d, gamma: %d\n",
   "Updated: ", remadd, dest->remoteport, dest->sbuf, dest->rbuf, dest->mode, dest->wad_ai, dest->wad_md, dest->curmss, dest->wad_maxssthresh, dest->wad_kaicnt, dest->wad_ifq, dest->divide, dest->floyd, dest->wad_vegas, dest->wad_alpha, dest->wad_beta, dest->wad_gamma);
          chg=0;
        }
	break;
      }
      olddest = dest;
    }
    if(!found) { 
      newdest = (struct Dest *) malloc (sizeof(struct Dest));
      if(newdest==NULL) {
        fprintf(stderr, "%s\n", "read_config: malloc error");
        return;
      }
      memcpy(newdest, destupd, sizeof(struct Dest));
      if(DestRoot==NULL) {
        olddest = DestRoot = newdest;
        olddest->next_dest = NULL; 
      }
      else {
        olddest->next_dest = newdest;
      }
/* for printing */
  net_addr.s_addr = htonl(newdest->remoteadd);
  remadd = inet_ntoa(net_addr);
  fprintf(stderr, "%s %s %d sbuf: %u rbuf: %u mode: %d wad_ai: %d wad_md: %d curmss: %d wad_maxssthresh: %d wad_kaicnt: %d wad_ifq: %d divide: %d, floyd: %d, vegas: %d, alpha: %d, beta: %d, gamma: %d\n",
   "Added: ", remadd, newdest->remoteport, newdest->sbuf, newdest->rbuf, newdest->mode, newdest->wad_ai, newdest->wad_md, newdest->curmss, newdest->wad_maxssthresh, newdest->wad_kaicnt, newdest->wad_ifq, newdest->divide, newdest->floyd, newdest->wad_vegas, newdest->wad_alpha, newdest->wad_beta, newdest->wad_gamma);
    }
/* for(dest=DestRoot; dest; dest=dest->next_dest)
  fprintf(stderr, "%u %u-->", dest->remoteadd, dest->remoteport);
fprintf(stderr, "\n");
*/
}


/* WalkProcWeb100() walks the directory tree comparing connections
   found with those in a linked list--those not found are passed to
   TuneConnections() to be added and possibly tuned.  Those found
   in the list are ignored as they have already been dealt with.
   Since this method has no built-in notification for the closing
   of a connection, the state of each connection in the list that is
   being tuned must be checked and, if a connection is in a closed 
   state, it is removed from tuning consideration.  Finally, the
   connections are removed from the linked list of active connections
   when it disappears from the directory, /proc/web100.
*/
void WalkProcWeb100(struct web100_agent *agent) {
  char path[32]="/proc/web100", errbuf[30];
  struct stat statbuf;
  struct dirent *dirp;
  DIR *dp;
  char *ptr;
  int dead=0, stage;
  struct active_cid *cidentry;
 
/* Check for closed connections */
    for(cidentry=CidRoot; cidentry; cidentry=cidentry->next_cid) {
      if(cidentry->stage == BUFFERSET) {
        stage = CheckState(cidentry, 0);
        if((stage==CIDCLOSED) && (cidentry->cid_dest)) {
	  UnlinkDestq(cidentry);
	  cidentry->stage = CIDCLOSED;
        }
      }
    }
/* Check for new connections in the /proc/web100 directory */  
  if((dp = opendir(path))==NULL) {
    err_sys("WalkProcWeb100: Can't read /proc/web100");
    exit(1);
  }
  ptr = path + strlen(path);
  *ptr++ = '/';
  *ptr = 0;
 
  while((dirp = readdir(dp)) != NULL) {
    if(strcmp(dirp->d_name, ".")==0 || strcmp(dirp->d_name, "..")==0)
      continue;
    strcpy(ptr, dirp->d_name);
/*If the directory entry is a valid directory */
    if(lstat(path, &statbuf) < 0) {
      sprintf(errbuf, "%s %s", "lstat error on ", path);
      err_sys(errbuf);
      continue;
    }
    if(S_ISDIR(statbuf.st_mode)) {
        cidentry = FindCid(atoi(dirp->d_name), 0);
        if(cidentry) {  /*is already in list of active connections */
          continue;
        }
        else {          /* not in active list--need to add */
          TuneConnections(agent, atoi(dirp->d_name)); 
        }
    *ptr = 0;
    }
  }
  closedir(dp);
  if(CidRoot) {        /* remove connections no longer in directory */
    dead = CleanUp();
  }
}

/* CleanUp()
   Traverses the linked list of active connections to
   clear out any connections that have gone away
*/

/*WHAT IF the first one is being freed???? What is CidRoot set to??*/

int CleanUp() {
  struct active_cid *next, *prev, *tmp;
  int dead=0;
 
  next = prev = CidRoot;
  while(next) {
    if(next->up == 0) {
/*fprintf(stderr, "UnlinkCid %d prev %d ", next->cid, prev->cid);*/
      UnlinkCid(next, prev);
      if(next->cid_dest)
	UnlinkDestq(next); 
      tmp=next;
      prev->next_cid = next = next->next_cid;
/*fprintf(stderr, "Free ");
fprintf(stderr, "%d...", tmp->cid);
*/
      free(tmp);
      dead++;
    }
    else {
      next->up = 0;
      prev = next;
      next = next->next_cid;
    }
  }
/*
if(dead > 0)
fprintf(stderr, "\n");
*/
  return(dead);
}

/* When a new connection is detected via a netlink socket or
   by checking the /proc/web100 directory, a web100 connection
   structure is set up and the connection end points and ports 
   are checked against those in the wad.conf file.  If a match 
   is found, the TCP state of the connection is checked and, 
   based on those results, web100 structures may be set up to 
   access the /proc/web100 files via the web100 interface and 
   tune the flow.  If the new connection does not pass those tests, 
   it is kept in a linked list so that it does not have to be dealt 
   with again until it closes.
*/
void TuneConnections(struct web100_agent *agent, int cid) {
  struct web100_connection *conn;
  struct active_cid *cidentry, *tmp;
  int err=0, ndx=0;
  char strcid[10];
  struct Dest *match;
  unsigned int tuned_sbuf=0, tuned_rbuf=0, new_floyd=0, new_ai=0, new_md=0, new_maxssthresh=0, new_kaicnt=0, new_mss=0, new_floydaimd=0, new_ifq=0;
  unsigned int new_vegas=0, new_alpha=0, new_beta=0, new_gamma=0;
  double cwndsegs=0.0;

    conn = web100_connection_lookup(agent, cid);
    cidentry = AddCid(conn); 
    match = FilterByConnection(conn);
    if(match) {
	cidentry->cid_dest = match;
	if(match->qfront == NULL)
	  match->qfront = cidentry;
	else 
	  match->qback->same_dest = cidentry;
	match->qback = cidentry;
        cidentry->stage = CheckState(cidentry, match->mode);
/* re-figure buffer sizes for all flows to the same destination
   when a new flow to that destination starts up */
        if(cidentry->stage==BUFFERSET){
          match->nconnects++;
          if(match->divide) {
	    tuned_sbuf = match->sbuf / match->nconnects;
	    tuned_rbuf = match->rbuf / match->nconnects;
	  }
	  else {
            tuned_sbuf = match->sbuf;
            tuned_rbuf = match->rbuf;
          }
	  cidentry->sbuf = tuned_sbuf;
	  cidentry->rbuf = tuned_rbuf;
/* make WAD adjustments at flow startup from data in config file */
	  if(match->wad_ifq >= 0) {
	    new_ifq = match->wad_ifq;
	    Web100RawWrite(WAD_IFQ, cidentry, &new_ifq);
	  }
	  if(match->wad_vegas >= 0) {
	    new_vegas = match->wad_vegas;
	    Web100RawWrite(WAD_Vegas, cidentry, &new_vegas);
	  }
	  if(match->wad_vegas > 0) {
	    if(match->wad_alpha) {
	      new_alpha = match->wad_alpha;
	      Web100RawWrite(WAD_VegasAlpha, cidentry, &new_alpha);
	    }
	    if(match->wad_beta) {
	      new_beta = match->wad_beta;
	      Web100RawWrite(WAD_VegasBeta, cidentry, &new_beta);
	    }
	    if(match->wad_gamma) {
	      new_gamma = match->wad_gamma;
	      Web100RawWrite(WAD_VegasGamma, cidentry, &new_gamma);
	    }
	  }
	  if(match->wad_maxssthresh) {
	    new_maxssthresh = match->wad_maxssthresh;
	    Web100RawWrite(WAD_MaxSsthresh, cidentry, &new_maxssthresh);
	  }
	  if(match->wad_kaicnt) {
	    new_kaicnt = match->wad_kaicnt;
	    Web100RawWrite(WAD_Kaicnt, cidentry, &new_kaicnt);
	  }
	  if(match->curmss) {
	    new_mss = match->curmss;
	    printf("Web100RawWrite for new_mss: %u\n", new_mss);
	    Web100RawWrite(CurMSStune, cidentry, &new_mss);
	  }
	  if(match->floyd==0) {
	    if(!(match->wad_ai==0 || match->wad_md==0)) {
	      new_ai = match->wad_ai;
	      new_md = match->wad_md;
	      RawWrite(cidentry->cid, &new_ai, &new_md);
	    } 
	  }
          else 
	   if(match->floyd==1) {
	          new_floydaimd = 1;
                  Web100RawWrite(WAD_FloydAIMD, cidentry, &new_floydaimd);
	   }
	   else 
	     if(match->floyd==3) {
	        ndx = cidentry->cwnd_ndx;
		cwndsegs=(double)cidentry->sbuf/1500;
	        floyd_aimd(cwndsegs, &ndx, &new_ai, &new_md);
	        RawWrite(cidentry->cid, &new_ai, &new_md);
	        cidentry->cwnd_ndx=ndx;
 	      }
	  
	  err=SetBuffers(cidentry);
	  if(!err)
fprintf(stderr, "NewBuffers       %d: sbuf %u rbuf %u\n",
  cidentry->cid, cidentry->sbuf, cidentry->rbuf);

	 if(match->divide) {
	  for(tmp=match->qfront; tmp; tmp=tmp->same_dest) {
	    if((tmp->stage == BUFFERSET) && (tmp != cidentry)) {
	      tmp->sbuf = tuned_sbuf;
	      tmp->rbuf = tuned_rbuf;
	      err=SetBuffers(tmp);
	      if(match->floyd==3) {
/* calculate ai & md based on new buffer size */
		ndx = tmp->cwnd_ndx;
		cwndsegs=(double)cidentry->sbuf/1500;
		floyd_aimd(cwndsegs, &ndx, &new_ai, &new_md);
	        if(ndx != tmp->cwnd_ndx) {
		  RawWrite(tmp->cid, &new_ai, &new_md);
	          tmp->cwnd_ndx=ndx;
		}
	      }
	      if(!err)
fprintf(stderr, "DecreasedBuffers %d: sbuf %u rbuf %u\n",
  tmp->cid, tmp->sbuf, tmp->rbuf);

            }
           }
	 }
/*
fprintf(stderr, "TuneConnections: ");
for(tmp=match->qfront; tmp; tmp=tmp->same_dest)
  printf("%d-->", tmp->cid);
printf("\n");
*/
      }
    }
    else {
      cidentry->stage = NOINTEREST;
    }
}

/*Adds a new connection to the linked list of active connections
*/
struct active_cid *AddCid(struct web100_connection *conn) {
  struct active_cid *new_cid, *next, *prev, *tmp;
 
  if(((struct active_cid *)new_cid = (struct active_cid *)malloc(sizeof(struct active_cid)))==NULL) {
    err_sys("AddCid: Out of memory");
    exit(1);
  }
  new_cid->cid = conn->cid;
  new_cid->stage = 0;
  new_cid->up = 1;
  new_cid->cwnd_ndx = 1;
  new_cid->same_dest = NULL;
  new_cid->cid_dest = NULL;
  new_cid->next_cid = CidRoot;
  CidRoot = new_cid; 
/*
tmp=CidRoot;
fprintf(stderr, "AddCid: ");
while(tmp){
  fprintf(stderr, "%d-->", tmp->cid);
  tmp=tmp->next_cid;
}
printf("\n");
*/
  return(new_cid);
}

/*Checks to see if a connection is in the linked list of active
    connections and, if so,  returns the address.  If the
    connection is being removed from the list, it is unlinked.
*/
struct active_cid *FindCid(int cid, int delete) {
  struct active_cid *tmp, *prev;
 
  prev = tmp = CidRoot;
  while(tmp) {
    if(tmp->cid == cid) {
/*fprintf(stderr, "FOUND %d in Stage %d\n", cid, tmp->stage);*/
      tmp->up = 1;
      if(delete)
	UnlinkCid(tmp, prev);
      return(tmp);
    }
    prev = tmp;
    tmp = tmp->next_cid;
  }
  return(struct active_cid *)(0);
}

int UnlinkCid(struct active_cid *tmp, struct active_cid *prev) {

  if(tmp == CidRoot)
    CidRoot = tmp->next_cid;
  else 
    prev->next_cid = tmp->next_cid;
/*  
fprintf(stderr, "UnlinkCid: %d Leaves:", tmp->cid);
for(tmp=CidRoot; tmp; tmp=tmp->next_cid)
  fprintf(stderr, "%d-->", tmp->cid);
printf("\n");
*/
}

/* Connection is closed--free memory:
	- adjust pointers and free cidentry structures
	- reapportion bufferspace to the rest of the connections
	  to the same destination
*/
int DeleteCid(int cid) {
  char strcid[10];
  struct active_cid *cidentry;

/* find the structure for this cid-- 
   if found, take it out of the active_cid linked list adjusting
   all pointers 
*/ 
  cidentry  = FindCid(cid, 1);
  if(cidentry) {
/* if it is connected with a Dest structure and is being tuned:
	- decrement the number of active flows for this destination
	- walk the Dest linked list re-allocating this flow's
	    bufferspace to the other active flows
   In any case: 
	- remove it from the Dest linked list adjusting all pointers
*/
    if(cidentry->cid_dest){
      UnlinkDestq(cidentry);
    }
/* free the connection and active_cid structures 
   associated with this cid */
    free(cidentry); 
    return(0);
  }
  else {
/* fprintf(stderr, "cid %d NOT Found--cannot delete!\n", 
  cidentry->cid); */
    return(1);
  }
}

/* UnlinkDestq() takes a connection out of the queue for a 
   particular destination and reallocates bufferspace for the
   remaining connections, if any.
*/
int UnlinkDestq(struct active_cid *cidentry) {
  struct Dest *dest = cidentry->cid_dest;
  struct active_cid *prev, *flow;
  unsigned int tuned_sbuf, tuned_rbuf;
  int realloc=0, err=0, ndx=0;
  int new_ai=0, new_md=0;
  double cwndsegs=0.0;

  if(cidentry->stage == BUFFERSET) {
/* if this flow is being tuned, decrement nconnects, 
   refigure and reallocate buffersizes
*/
    if(--dest->nconnects) 
      if(dest->divide) {
        tuned_sbuf = dest->sbuf / dest->nconnects;
        tuned_rbuf = dest->rbuf / dest->nconnects;
        realloc = 1;
      }
  } 
  prev = flow = dest->qfront;
  while(flow) {
    if(flow == cidentry) {
/*fprintf(stderr, "CID %d has terminated\n", flow->cid);*/
      if(flow == dest->qfront)
        dest->qfront = flow->same_dest;
      else if(flow == dest->qback)
	dest->qback = prev;
      prev->same_dest = flow->same_dest;
    }
    else {
      if(flow->stage == BUFFERSET && realloc) {
	flow->sbuf = tuned_sbuf;
	flow->rbuf = tuned_rbuf; 
	err=SetBuffers(flow);
	if(dest->floyd==3) {
/* calculate ai & md based on new buffer size */
	  ndx = flow->cwnd_ndx;
	  cwndsegs=(double)flow->sbuf/1500;
	  floyd_aimd(cwndsegs, &ndx, &new_ai, &new_md);
	  if(ndx != flow->cwnd_ndx) {
	    RawWrite(flow->cid, &new_ai, &new_md);
	    flow->cwnd_ndx=ndx;
	  }
	}
	if(!err) {
	  if(dest->floyd==0) {
	    new_ai = dest->wad_ai;
	    new_md = dest->wad_md;
	  }
fprintf(stderr, "IncreasedBuffers %d: sbuf %u rbuf %u\n",
flow->cid, flow->sbuf, flow->rbuf);
	}

      }
    }
    prev = flow;
    flow = flow->same_dest;
  }
  cidentry->cid_dest = NULL;

/*for(flow=dest->qfront; flow; flow=flow->same_dest)
  printf("%s-->", flow->cid);
printf("\n");
*/
}

/* FilterByConnection:
   Filters Out: unwanted ports 
 	        connections not included in the config file, wad.conf
*/
struct Dest *FilterByConnection(struct web100_connection *conn) {
  struct Dest *dest = DestRoot;
  int i=0, aok=0;
  u_int32_t dst_addr;

  if((conn->spec.dst_port == HTTP_PORT) ||
     (conn->spec.src_port == BWSERVER) ||
     (conn->spec.dst_port == GRID_CTL) ||
     (conn->spec.dst_port == MIB_CTL) ||
     (conn->spec.dst_port == SSH_PORT) ||
     (conn->spec.src_port == SSH_PORT)) { 
    return(0);
  }

  while(dest) {
    dst_addr = ntohl(conn->spec.dst_addr);
    dst_addr = dst_addr & dest->remotemsk;
    if(dst_addr == dest->remoteadd) {
      if(dest->remoteport > 0)
	if(conn->spec.dst_port == dest->remoteport) {
	  if(aok = LocalCheck(conn, dest))
	    break;
	}
	else
	  continue;
      else {
	if(aok = LocalCheck(conn, dest))
	  break;
      }
    }
    else
      dest = dest->next_dest;
  } 
  if(aok) {
      return(dest);
  }
  else {
    return(0);
  }
}

int LocalCheck(struct web100_connection *conn, struct Dest *dest) {
  u_int32_t src_addr;

  if(dest->localadd > 0) {
    src_addr = ntohl(conn->spec.src_addr);
    src_addr = src_addr & dest->localmsk;
    if(dest->localadd == src_addr)
      if(dest->localport > 0)
        if(conn->spec.src_port == dest->localport)
	  return(1);
        else
	  return(0);
      else
	return(1);
    else
      return(0);
  }
  else
    return(1);
}


/*gets the updated State of the connection via the web100 interface 
  (TCP states are redefined in web100_stats.h)
    If State == 3, 4 or 5 and the application has not used
	setsockopt() OR you want to force tuning anyhow:  
		    - stage = BUFFERSET
    else	    - stage = NOINTEREST
    If State == 1, 2, or 6-12 the connection is closing:
		    - state = CIDCLOSED
*/
int CheckState(struct active_cid *cidentry, int mode) {
  unsigned int state=0, sauto=0, rauto=0;
  unsigned char buf[8], errbuf[30];
  
  if((Web100RawRead(CidState, cidentry, buf)) == 0) {   
    state = *(u_int16_t *)buf; 
    if((state > 2) && (state < 6)) { 
      Web100RawRead(X_RBufMode, cidentry, buf);
      rauto = *(u_int16_t *)buf;
      Web100RawRead(X_SBufMode, cidentry, buf);
      sauto = *(u_int16_t *)buf;
      if(mode) {
	if(rauto>0) {
	  rauto=0;
	  Web100RawWrite(X_RBufMode, cidentry, &rauto);
	}
        if(sauto > 0) {
	  sauto = 0;
	  Web100RawWrite(X_SBufMode, cidentry, &sauto);
	}
	return(BUFFERSET);
      }
      else { /* need linux auto-tuning */
        if(sauto < 1) {
	  sauto = 1;
	  Web100RawWrite(X_SBufMode, cidentry, &sauto);
	}
	if(rauto < 1) {
	  rauto=1;
	  Web100RawWrite(X_RBufMode, cidentry, &rauto);
	}
        return(NOINTEREST);
      }
    }
    else {
      return(CIDCLOSED);
    }
  }
  else {
    sprintf(errbuf, "%s %s\n", "CheckState: Couldnot get State for", cidentry->cid); 
    err_sys(errbuf);
    return(NOINTEREST);
  }
}

/* SetBuffers():
   Attempts to set the buffer sizes for a particular tcp
   connection based on the sizes contained in the wad.conf file.
*/
int SetBuffers(struct active_cid *cidentry) {
  int fd, length;
  u_int32_t mode;
  unsigned int sbuf, rbuf;
  char filename[64];

/***Should check here to make sure buffers read from config file
    were greater than 0--before resetting values***/

  sprintf(filename, "%s/%d/%s", "/proc/web100", cidentry->cid, LimCwnd->group->name);
  if((fd=open(filename, O_RDWR))<0) {
    err_sys("SetBuffers: file open");
    return(-1);
  }
  rbuf = cidentry->rbuf;
  if((lseek(fd, LimRwin->offset, SEEK_SET)) < 0) {
    err_sys("SetBuffers: lseek LimRwin");
    close(fd);
    return(-1);
  }
  if((write(fd, &rbuf, LimRwin->len)) != LimRwin->len) {
    err_sys("SetBuffers: write LimRwin");
    close(fd);
    return(-1);
  }

  sbuf = cidentry->sbuf;
  if((lseek(fd, LimCwnd->offset, SEEK_SET)) < 0) {
    err_sys("SetBuffers: lseek LimCwnd");
    close(fd);
    return(-1);
  }
  if((write(fd, &sbuf, LimCwnd->len)) != LimCwnd->len) {
    err_sys("SetBuffers: write LimCwnd");
    close(fd);
    return(-1);
  }
  return(0);
}

/*
Modified from web100.c:
web100_raw_read - read a variable from a connection into a buffer
*/
int Web100RawRead(web100_var *var, struct active_cid *cidentry, void *buf)
{
  int fd;  
  char filename[64];
 
  sprintf(filename, "%s/%d/%s", "/proc/web100", cidentry->cid, var->group->name);
  if ((fd = open(filename, O_RDONLY)) < 0) {
    perror("Web100RawRead: fopen");
    exit(1);
  }
  if ((lseek(fd, var->offset, SEEK_SET)) < 0) {
    perror("Web100RawRead: lseek");
    exit(1);
  }
  if ((read(fd, buf, var->len)) != var->len) {
    perror("Web100RawRead: read");
    exit(1);
  }
  if (close(fd))
    perror("Web100RawRead: close");
 
  return(0);
 
}

/*
Modified from web100.c:
web100_raw_write - write a variable into a connection from a buffer
QUESTION!!--too many opens?!
*/
int Web100RawWrite(struct web100_var *var, struct active_cid *cidentry, void *buf) {
  int fd;  
  char filename[64];
 
  sprintf(filename, "%s/%d/%s", "/proc/web100", cidentry->cid, var->group->name);
  if ((fd = open(filename, O_RDWR)) < 0) {
    perror("Web100RawWrite: fopen");
    exit(1);
  }
  if ((lseek(fd, var->offset, SEEK_SET)) < 0) {
    perror("Web100RawWrite: lseek");
    exit(1);
  }
  if ((write(fd, buf, var->len)) != var->len) {
    perror("Web100RawWrite: write");
    exit(1);
  }
  if (close(fd))
    perror("Web100RawWrite: close");
 
  return(0);
}


/**********DYNAMIC FLOYD AIMD***************************************/

int RawRead(int cid, u_int32_t *currentcwnd, u_int32_t *currentmss){
  int fd;
  char filename[64];
  u_int32_t curcwnd=0, curmss=0;
 
  sprintf(filename, "%s%d/%s", WEB100_ROOT_DIR, cid, "read");
  if ((fd = open(filename, O_RDONLY)) < 0) {
      printf("%s: OPEN FAILURE\n", filename);
      exit(1);
  }
  if ((lseek(fd, CurCwnd->offset, SEEK_SET)) < 0) {
      printf("lseek FAILURE for CurCwnd\n");
      exit(1);
  }
  if ((read(fd, &curcwnd, CurCwnd->len)) < 0 ) {
      printf("read FAILURE for CurCwnd\n");
      exit(1);
  }
  if ((lseek(fd, CurMSS->offset, SEEK_SET)) < 0) {
      printf("lseek FAILURE for CurMSS\n");
      exit(1);
  }
  if ((read(fd, &curmss, CurMSS->len)) < 0 ) {
      printf("read FAILURE for CurMSS\n");
      exit(1);
  }
  if (close(fd))
      perror("close FAILURE\n");
  *(currentcwnd) = curcwnd;
  *(currentmss) = curmss;

  return(0);
}
 
int RawWrite(int cid, int *wad_ai, int *wad_md) {  
  char filename[64];
  int fd;
 
  sprintf(filename, "%s%d/%s", WEB100_ROOT_DIR, cid, "tune");
  if ((fd = open(filename, O_RDWR)) < 0) {
      printf("%s: OPEN FAILURE\n", filename);
      exit(1);
  }
  if ((lseek(fd, WAD_MD->offset, SEEK_SET)) < 0) {
      printf("%s: fseek FAILURE\n", filename);
      exit(1);
  }
  if ((write(fd, wad_md, WAD_MD->len)) != WAD_MD->len) {
      printf("%s: write FAILURE\n", filename);
      exit(1);
  }
  if ((lseek(fd, WAD_AI->offset, SEEK_SET)) < 0) {
      printf("%s: fseek FAILURE\n", filename);
      exit(1);
  }
  if ((write(fd, wad_ai, WAD_AI->len)) != WAD_AI->len) {
      printf("%s: fwrite FAILURE\n", filename);
      exit(1);
  }
  if (close(fd))
      perror("web100_raw_write: fclose");

  return(0);
}
 
/* floyd aimd calculator  */
static struct Aimd_Vals {
        unsigned int cwnd;
        unsigned int increment;
        unsigned int decrement;
} aimd_vals[] = {
     0,     8, /*     1 */   128, /*  0.50 */
    38,     8, /*     1 */   128, /*  0.50 */
   118,    16, /*     2 */   112, /*  0.44 */
   221,    24, /*     3 */   104, /*  0.41 */
   347,    32, /*     4 */    98, /*  0.38 */
   495,    40, /*     5 */    93, /*  0.37 */
   663,    48, /*     6 */    89, /*  0.35 */
   851,    56, /*     7 */    86, /*  0.34 */
  1058,    64, /*     8 */    83, /*  0.33 */
  1284,    72, /*     9 */    81, /*  0.32 */
  1529,    80, /*    10 */    78, /*  0.31 */
  1793,    88, /*    11 */    76, /*  0.30 */
  2076,    96, /*    12 */    74, /*  0.29 */
  2378,   104, /*    13 */    72, /*  0.28 */
  2699,   112, /*    14 */    71, /*  0.28 */
  3039,   120, /*    15 */    69, /*  0.27 */
  3399,   128, /*    16 */    68, /*  0.27 */
  3778,   136, /*    17 */    66, /*  0.26 */
  4177,   144, /*    18 */    65, /*  0.26 */
  4596,   152, /*    19 */    64, /*  0.25 */
  5036,   160, /*    20 */    62, /*  0.25 */
  5497,   168, /*    21 */    61, /*  0.24 */
  5979,   176, /*    22 */    60, /*  0.24 */
  6483,   184, /*    23 */    59, /*  0.23 */
  7009,   192, /*    24 */    58, /*  0.23 */
  7558,   200, /*    25 */    57, /*  0.22 */
  8130,   208, /*    26 */    56, /*  0.22 */
  8726,   216, /*    27 */    55, /*  0.22 */
  9346,   224, /*    28 */    54, /*  0.21 */
  9991,   232, /*    29 */    53, /*  0.21 */
 10661,   240, /*    30 */    52, /*  0.21 */
 11358,   248, /*    31 */    52, /*  0.20 */
 12082,   256, /*    32 */    51, /*  0.20 */
 12834,   264, /*    33 */    50, /*  0.20 */
 13614,   272, /*    34 */    49, /*  0.19 */
 14424,   280, /*    35 */    48, /*  0.19 */
 15265,   288, /*    36 */    48, /*  0.19 */
 16137,   296, /*    37 */    47, /*  0.19 */
 17042,   304, /*    38 */    46, /*  0.18 */
 17981,   312, /*    39 */    45, /*  0.18 */
 18955,   320, /*    40 */    45, /*  0.18 */
 19965,   328, /*    41 */    44, /*  0.17 */
 21013,   336, /*    42 */    43, /*  0.17 */
 22101,   344, /*    43 */    43, /*  0.17 */
 23230,   352, /*    44 */    42, /*  0.17 */
 24402,   360, /*    45 */    41, /*  0.16 */
 25618,   368, /*    46 */    41, /*  0.16 */
 26881,   376, /*    47 */    40, /*  0.16 */
 28193,   384, /*    48 */    39, /*  0.16 */
 29557,   392, /*    49 */    39, /*  0.15 */
 30975,   400, /*    50 */    38, /*  0.15 */
 32450,   408, /*    51 */    38, /*  0.15 */
 33986,   416, /*    52 */    37, /*  0.15 */
 35586,   424, /*    53 */    36, /*  0.14 */
 37253,   432, /*    54 */    36, /*  0.14 */
 38992,   440, /*    55 */    35, /*  0.14 */
 40808,   448, /*    56 */    35, /*  0.14 */
 42707,   456, /*    57 */    34, /*  0.13 */
 44694,   464, /*    58 */    33, /*  0.13 */
 46776,   472, /*    59 */    33, /*  0.13 */
 48961,   480, /*    60 */    32, /*  0.13 */
 51258,   488, /*    61 */    32, /*  0.13 */
 53677,   496, /*    62 */    31, /*  0.12 */
 56230,   504, /*    63 */    30, /*  0.12 */
 58932,   512, /*    64 */    30, /*  0.12 */
 61799,   520, /*    65 */    29, /*  0.12 */
 64851,   528, /*    66 */    28, /*  0.11 */
 68113,   536, /*    67 */    28, /*  0.11 */
 71617,   544, /*    68 */    27, /*  0.11 */
 75401,   552, /*    69 */    26, /*  0.10 */
 79517,   560, /*    70 */    26, /*  0.10 */
 84035,   568, /*    71 */    25, /*  0.10 */
 89053,   576, /*    72 */    24, /*  0.10 */
 94717,   584, /*    73 */    23, /*  0.09 */
999999,   584, /*    73 */    23  /*  0.09 */
};
 
floyd_aimd( double cwndseg, int *tblndx, int *wadai, int *wadmd) {
    int tmpndx;
    double multiplier;
/* fprintf(stderr, "floyd_aimd: cwndseg: %f  ndx: %d \n", cwndseg, *(tblndx)); */
        tmpndx = *(tblndx);
        if (cwndseg >  aimd_vals[tmpndx].cwnd ){
                /* find new upper bound */
                while (cwndseg >  aimd_vals[tmpndx].cwnd ) tmpndx++;
        } else  if (cwndseg < aimd_vals[tmpndx-1].cwnd ){
                /* find new lower bound */
                while (cwndseg < aimd_vals[tmpndx-1].cwnd ) tmpndx--;
        } else return;   /* no change */
/* fprintf(stderr,"floyd_aimd: %f  tmpndx %d ai:%d md: %d\n", cwndseg,tmpndx,aimd_vals[tmpndx].increment, aimd_vals[tmpndx].decrement); */
        *(wadai) = aimd_vals[tmpndx].increment;
        *(wadmd) = aimd_vals[tmpndx].decrement;
        *(tblndx) = tmpndx;
}
 
/* MonitorCurrentCwnd()
   Following the linked list of destinations read from the wad.conf,
   if a destination has floyd==2 (means do floyd aimd calculations
   every .1 second) then, in turn, check each active flow to this
   destination--read CurrentCwnd from the /proc/web100/cid/read file
   and, if the value has changed so that a new additive
   increase/multiplicative decrease is needed, retrieve values for
   WAD_AI and WAD_MD from a table and write them to the
   /proc/web100/cid/tune file.
*/
 
void *MonitorCurrentCwnd() {
  unsigned int currentcwnd=0, currentmss=0;
  unsigned char buf[8], errbuf[30], message[512];
  struct Dest *tmpdest;
  struct active_cid *tmpcid;
  int fd, tmpndx, wadai, wadmd, n;
  struct timeval tv;
  double naptime = 0.1;
 
  pthread_detach(pthread_self());
  while(1) {
    for(tmpdest=DestRoot; tmpdest; tmpdest=tmpdest->next_dest) {
      if(tmpdest->floyd == 2) {
        for(tmpcid=tmpdest->qfront; tmpcid; tmpcid=tmpcid->same_dest){
          if(tmpcid->stage==BUFFERSET) {
/* read for CurrentCwnd, check to see if it has significantly changed
   and, if so, get new values for WAD_AI and WAD_MD and write them
   to the /tune file
*/
            RawRead(tmpcid->cid, &currentcwnd, &currentmss);
            tmpndx = tmpcid->cwnd_ndx;
            floyd_aimd(currentcwnd/currentmss, &tmpndx, &wadai, &wadmd);
            if(tmpndx != tmpcid->cwnd_ndx) {
              RawWrite(tmpcid->cid, &wadai, &wadmd);
              tmpcid->cwnd_ndx = tmpndx;
fprintf(stderr, "Cid: %d cwnd: %u(%d) wadai: %d wadmd: %d ndx: %d\n",
tmpcid->cid, currentcwnd, currentcwnd/currentmss, wadai, wadmd, tmpndx);
 
            }
          }
        }
      }
    }
    tv.tv_sec = 0;
    tv.tv_usec = (naptime * 1000000);
    n = select((int)NULL, NULL, NULL, NULL, &tv);
  }
  return(NULL);
}

